@@ -5,12 +5,25 @@ import torch.nn.functional as F
 from einops import rearrange, repeat
 from einops.layers.torch import Rearrange
 
-from torch_geometric.nn import MessagePassing
-
 # types
 
 from typing import Optional, List, Union
-from torch_geometric.typing import Adj, Size, OptTensor, Tensor
+
+# pytorch geometric
+
+try:
+    from torch_geometric.nn import MessagePassing
+    from torch_geometric.typing import Adj, Size, OptTensor, Tensor
+    PYG_AVAILABLE = True
+except:
+    Tensor = OptTensor = Adj = MessagePassing = Size = object
+    PYG_AVAILABLE = False
+    
+    # to stop throwing errors from type suggestions
+    Adj = object
+    Size = object
+    OptTensor = object
+    Tensor = object
 
 # helper functions
 
@@ -50,7 +63,7 @@ def fourier_encode_dist(x, num_encodings = 4, include_self = True):
 def embedd_token(x, dims, layers):
     stop_concat = -len(dims)
     to_embedd = x[:, stop_concat:].long()
-    for i,emb_layer in enumerate(layer):
+    for i,emb_layer in enumerate(layers):
         # the portion corresponding to `to_embedd` part gets dropped
         x = torch.cat([ x[:, :stop_concat], 
                         emb_layer( to_embedd[:, i] ) 
@@ -72,16 +85,83 @@ SiLU = nn.SiLU if hasattr(nn, 'SiLU') else Swish_
 # https://github.com/lucidrains/se3-transformer-pytorch/blob/main/se3_transformer_pytorch/se3_transformer_pytorch.py#L95
 
 class CoorsNorm(nn.Module):
-    def __init__(self, eps = 1e-8):
+    def __init__(self, eps = 1e-8, scale_init = 1.):
         super().__init__()
         self.eps = eps
-        self.fn = nn.Sequential(nn.LayerNorm(1), nn.GELU())
+        scale = torch.zeros(1).fill_(scale_init)
+        self.scale = nn.Parameter(scale)
 
     def forward(self, coors):
         norm = coors.norm(dim = -1, keepdim = True)
         normed_coors = coors / norm.clamp(min = self.eps)
-        phase = self.fn(norm)
-        return (phase * normed_coors)
+        return normed_coors * self.scale
+
+# global linear attention
+
+class Attention(nn.Module):
+    def __init__(self, dim, heads = 8, dim_head = 64):
+        super().__init__()
+        inner_dim = heads * dim_head
+        self.heads = heads
+        self.scale = dim_head ** -0.5
+
+        self.to_q = nn.Linear(dim, inner_dim, bias = False)
+        self.to_kv = nn.Linear(dim, inner_dim * 2, bias = False)
+        self.to_out = nn.Linear(inner_dim, dim)
+
+    def forward(self, x, context, mask = None):
+        h = self.heads
+
+        q = self.to_q(x)
+        kv = self.to_kv(context).chunk(2, dim = -1)
+
+        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (q, *kv))
+        dots = einsum('b h i d, b h j d -> b h i j', q, k) * self.scale
+
+        if exists(mask):
+            mask_value = -torch.finfo(dots.dtype).max
+            mask = rearrange(mask, 'b n -> b () () n')
+            dots.masked_fill_(~mask, mask_value)
+
+        attn = dots.softmax(dim = -1)
+        out = einsum('b h i j, b h j d -> b h i d', attn, v)
+
+        out = rearrange(out, 'b h n d -> b n (h d)', h = h)
+        return self.to_out(out)
+
+class GlobalLinearAttention(nn.Module):
+    def __init__(
+        self,
+        *,
+        dim,
+        heads = 8,
+        dim_head = 64
+    ):
+        super().__init__()
+        self.norm_seq = nn.LayerNorm(dim)
+        self.norm_queries = nn.LayerNorm(dim)
+        self.attn1 = Attention(dim, heads, dim_head)
+        self.attn2 = Attention(dim, heads, dim_head)
+
+        self.ff = nn.Sequential(
+            nn.LayerNorm(dim),
+            nn.Linear(dim, dim * 4),
+            nn.GELU(),
+            nn.Linear(dim * 4, dim)
+        )
+
+    def forward(self, x, queries, mask = None):
+        res_x, res_queries = x, queries
+        x, queries = self.norm_seq(x), self.norm_queries(queries)
+
+        induced = self.attn1(queries, x, mask = mask)
+        out     = self.attn2(x, induced)
+
+        x =  out + res_x
+        queries = induced + res_queries
+
+        x = self.ff(x) + x
+        return x, queries
 
 # classes
 
@@ -92,18 +172,19 @@ class EGNN(nn.Module):
         edge_dim = 0,
         m_dim = 16,
         fourier_features = 0,
-        norm_rel_coors = False,
-        norm_coor_weights = False,
         num_nearest_neighbors = 0,
         dropout = 0.0,
         init_eps = 1e-3,
         norm_feats = False,
         norm_coors = False,
+        norm_coors_scale_init = 1e-2,
         update_feats = True,
         update_coors = True,
         only_sparse_neighbors = False,
         valid_radius = float('inf'),
-        m_pool_method = 'sum'
+        m_pool_method = 'sum',
+        soft_edges = False,
+        coor_weights_clamp_value = None
     ):
         super().__init__()
         assert m_pool_method in {'sum', 'mean'}, 'pool method must be either sum or mean'
@@ -122,8 +203,13 @@ class EGNN(nn.Module):
             SiLU()
         )
 
+        self.edge_gate = nn.Sequential(
+            nn.Linear(m_dim, 1),
+            nn.Sigmoid()
+        ) if soft_edges else None
+
         self.node_norm = nn.LayerNorm(dim) if norm_feats else nn.Identity()
-        self.coors_norm = CoorsNorm() if norm_coors else nn.Identity()
+        self.coors_norm = CoorsNorm(scale_init = norm_coors_scale_init) if norm_coors else nn.Identity()
 
         self.m_pool_method = m_pool_method
 
@@ -134,12 +220,6 @@ class EGNN(nn.Module):
             nn.Linear(dim * 2, dim),
         ) if update_feats else None
 
-        self.norm_coor_weights = norm_coor_weights
-        self.norm_rel_coors = norm_rel_coors
-
-        if norm_rel_coors:
-            self.rel_coors_scale = nn.Parameter(torch.ones(1))
-
         self.coors_mlp = nn.Sequential(
             nn.Linear(m_dim, m_dim * 4),
             dropout,
@@ -151,6 +231,8 @@ class EGNN(nn.Module):
         self.only_sparse_neighbors = only_sparse_neighbors
         self.valid_radius = valid_radius
 
+        self.coor_weights_clamp_value = coor_weights_clamp_value
+
         self.init_eps = init_eps
         self.apply(self.init_)
 
@@ -165,7 +247,7 @@ class EGNN(nn.Module):
         if exists(mask):
             num_nodes = mask.sum(dim = -1)
 
-        use_nearest = num_nearest > 0
+        use_nearest = num_nearest > 0 or only_sparse_neighbors
 
         rel_coors = rearrange(coors, 'b i d -> b i () d') - rearrange(coors, 'b j d -> b () j d')
         rel_dist = (rel_coors ** 2).sum(dim = -1, keepdim = True)
@@ -173,15 +255,15 @@ class EGNN(nn.Module):
         i = j = n
 
         if use_nearest:
-            ranking = rel_dist[..., 0]
+            ranking = rel_dist[..., 0].clone()
 
             if exists(mask):
-                rank_mask = mask[:, None, :] * mask[:, None, :]
+                rank_mask = mask[:, :, None] * mask[:, None, :]
                 ranking.masked_fill_(~rank_mask, 1e5)
 
             if exists(adj_mat):
                 if len(adj_mat.shape) == 2:
-                    adj_mat = repeat(adj_mat, 'i j -> b i j', b = b)
+                    adj_mat = repeat(adj_mat.clone(), 'i j -> b i j', b = b)
 
                 if only_sparse_neighbors:
                     num_nearest = int(adj_mat.float().sum(dim = -1).max().item())
@@ -189,8 +271,8 @@ class EGNN(nn.Module):
 
                 self_mask = rearrange(torch.eye(n, device = device, dtype = torch.bool), 'i j -> () i j')
 
+                adj_mat = adj_mat.masked_fill(self_mask, False)
                 ranking.masked_fill_(self_mask, -1.)
-                adj_mat.masked_fill_(self_mask, False)
                 ranking.masked_fill_(adj_mat, 0.)
 
             nbhd_ranking, nbhd_indices = ranking.topk(num_nearest, dim = -1, largest = False)
@@ -224,6 +306,9 @@ class EGNN(nn.Module):
 
         m_ij = self.edge_mlp(edge_input)
 
+        if exists(self.edge_gate):
+            m_ij = m_ij * self.edge_gate(m_ij)
+
         if exists(mask):
             mask_i = rearrange(mask, 'b i -> b i ()')
 
@@ -238,21 +323,23 @@ class EGNN(nn.Module):
             coor_weights = self.coors_mlp(m_ij)
             coor_weights = rearrange(coor_weights, 'b i j () -> b i j')
 
-            if self.norm_coor_weights:
-                coor_weights = coor_weights.tanh()
-
             rel_coors = self.coors_norm(rel_coors)
 
             if exists(mask):
                 coor_weights.masked_fill_(~mask, 0.)
 
+            if exists(self.coor_weights_clamp_value):
+                clamp_value = self.coor_weights_clamp_value
+                coor_weights.clamp_(min = -clamp_value, max = clamp_value)
+
             coors_out = einsum('b i j, b i j c -> b i c', coor_weights, rel_coors) + coors
         else:
             coors_out = coors
 
         if exists(self.node_mlp):
-            m_ij_mask = rearrange(mask, '... -> ... ()')
-            m_ij = m_ij.masked_fill(~m_ij_mask, 0.)
+            if exists(mask):
+                m_ij_mask = rearrange(mask, '... -> ... ()')
+                m_ij = m_ij.masked_fill(~m_ij_mask, 0.)
 
             if self.m_pool_method == 'mean':
                 if exists(mask):
@@ -281,26 +368,90 @@ class EGNN_Network(nn.Module):
         dim,
         num_tokens = None,
         num_edge_tokens = None,
+        num_positions = None,
         edge_dim = 0,
+        num_adj_degrees = None,
+        adj_dim = 0,
+        global_linear_attn_every = 0,
+        global_linear_attn_heads = 8,
+        global_linear_attn_dim_head = 64,
+        num_global_tokens = 4,
         **kwargs
     ):
         super().__init__()
+        assert not (exists(num_adj_degrees) and num_adj_degrees < 1), 'make sure adjacent degrees is greater than 1'
+        self.num_positions = num_positions
+
         self.token_emb = nn.Embedding(num_tokens, dim) if exists(num_tokens) else None
+        self.pos_emb = nn.Embedding(num_positions, dim) if exists(num_positions) else None
         self.edge_emb = nn.Embedding(num_edge_tokens, edge_dim) if exists(num_edge_tokens) else None
+        self.has_edges = edge_dim > 0
+
+        self.num_adj_degrees = num_adj_degrees
+        self.adj_emb = nn.Embedding(num_adj_degrees + 1, adj_dim) if exists(num_adj_degrees) and adj_dim > 0 else None
+
+        edge_dim = edge_dim if self.has_edges else 0
+        adj_dim = adj_dim if exists(num_adj_degrees) else 0
+
+        has_global_attn = global_linear_attn_every > 0
+        self.global_tokens = None
+        if has_global_attn:
+            self.global_tokens = nn.Parameter(torch.randn(num_global_tokens, dim))
 
         self.layers = nn.ModuleList([])
-        for _ in range(depth):
-            self.layers.append(EGNN(dim = dim, edge_dim = edge_dim, norm_feats = True, **kwargs))
+        for ind in range(depth):
+            is_global_layer = has_global_attn and (ind % global_linear_attn_every) == 0
+
+            self.layers.append(nn.ModuleList([
+                GlobalLinearAttention(dim = dim, heads = global_linear_attn_heads, dim_head = global_linear_attn_dim_head) if is_global_layer else None,
+                EGNN(dim = dim, edge_dim = (edge_dim + adj_dim), norm_feats = True, **kwargs),
+            ]))
 
     def forward(self, feats, coors, adj_mat = None, edges = None, mask = None):
+        b, device = feats.shape[0], feats.device
+
         if exists(self.token_emb):
             feats = self.token_emb(feats)
 
+        if exists(self.pos_emb):
+            n = feats.shape[1]
+            assert n <= self.num_positions, f'given sequence length {n} must be less than the number of positions {self.num_positions} set at init'
+            pos_emb = self.pos_emb(torch.arange(n, device = device))
+            feats += rearrange(pos_emb, 'n d -> () n d')
+
         if exists(edges) and exists(self.edge_emb):
             edges = self.edge_emb(edges)
 
-        for layer in self.layers:
-            feats, coors = layer(feats, coors, adj_mat = adj_mat, edges = edges, mask = mask)
+        # create N-degrees adjacent matrix from 1st degree connections
+        if exists(self.num_adj_degrees):
+            assert exists(adj_mat), 'adjacency matrix must be passed in (keyword argument adj_mat)'
+
+            if len(adj_mat.shape) == 2:
+                adj_mat = repeat(adj_mat.clone(), 'i j -> b i j', b = b)
+
+            adj_indices = adj_mat.clone().long()
+
+            for ind in range(self.num_adj_degrees - 1):
+                degree = ind + 2
+
+                next_degree_adj_mat = (adj_mat.float() @ adj_mat.float()) > 0
+                next_degree_mask = (next_degree_adj_mat.float() - adj_mat.float()).bool()
+                adj_indices.masked_fill_(next_degree_mask, degree)
+                adj_mat = next_degree_adj_mat.clone()
+
+            if exists(self.adj_emb):
+                adj_emb = self.adj_emb(adj_indices)
+                edges = torch.cat((edges, adj_emb), dim = -1) if exists(edges) else adj_emb
+
+        global_tokens = None
+        if exists(self.global_tokens):
+            global_tokens = repeat(self.global_tokens, 'n d -> b n d', b = b)
+
+        for global_attn, egnn in self.layers:
+            if exists(global_attn):
+                feats, global_tokens = global_attn(feats, global_tokens, mask = mask)
+
+            feats, coors = egnn(feats, coors, adj_mat = adj_mat, edges = edges, mask = mask)
 
         return feats, coors
 
@@ -308,6 +459,7 @@ class EGNN_sparse(MessagePassing):
     """ Different from the above since it separates the edge assignment
         from the computation (this allows for great reduction in time and 
         computations when the graph is locally or sparse connected).
+        * aggr: one of ["add", "mean", "max"]
     """
     def __init__(
         self,
@@ -316,72 +468,85 @@ class EGNN_sparse(MessagePassing):
         edge_attr_dim = 0,
         m_dim = 16,
         fourier_features = 0,
-        update_coors = True, 
-        update_feats = True,
+        soft_edge = 0,
         norm_feats = False,
-        norm_rel_coors = False,
-        norm_coor_weights = False,
+        norm_coors = False,
+        update_feats = True,
+        update_coors = True, 
         dropout = 0.,
-        init_eps = 1e-3
+        coor_weights_clamp_value = None, 
+        aggr = "add",
+        **kwargs
     ):
-        super().__init__()
+        assert aggr in {'add', 'sum', 'max', 'mean'}, 'pool method must be a valid option'
+        assert update_feats or update_coors, 'you must update either features, coordinates, or both'
+        kwargs.setdefault('aggr', aggr)
+        super(EGNN_sparse, self).__init__(**kwargs)
         # model params
         self.fourier_features = fourier_features
         self.feats_dim = feats_dim
         self.pos_dim = pos_dim
+        self.m_dim = m_dim
+        self.soft_edge = soft_edge
         self.norm_feats = norm_feats
+        self.norm_coors = norm_coors
         self.update_coors = update_coors
         self.update_feats = update_feats
+        self.coor_weights_clamp_value = None
 
-        edge_input_dim = (fourier_features * 2) + (feats_dim * 2) + edge_attr_dim + 1
-        dropout = nn.Dropout(dropout) if dropout > 0 else nn.Identity()
+        self.edge_input_dim = (fourier_features * 2) + edge_attr_dim + 1 + (feats_dim * 2)
+        self.dropout = nn.Dropout(dropout) if dropout > 0 else nn.Identity()
 
         # EDGES
         self.edge_mlp = nn.Sequential(
-            nn.Linear(edge_input_dim, edge_input_dim * 2),
-            dropout,
+            nn.Linear(self.edge_input_dim, self.edge_input_dim * 2),
+            self.dropout,
             SiLU(),
-            nn.Linear(edge_input_dim * 2, m_dim),
+            nn.Linear(self.edge_input_dim * 2, m_dim),
             SiLU()
         )
 
+        self.edge_weight = nn.Sequential(nn.Linear(m_dim, 1), 
+                                         nn.Sigmoid()
+        ) if soft_edge else None
+
         # NODES
         self.node_norm = nn.LayerNorm(feats_dim) if norm_feats else nn.Identity()
+        self.coors_norm = CoorsNorm() if norm_coors else nn.Identity()
 
         self.node_mlp = nn.Sequential(
             nn.Linear(feats_dim + m_dim, feats_dim * 2),
-            dropout,
+            self.dropout,
             SiLU(),
             nn.Linear(feats_dim * 2, feats_dim),
         ) if update_feats else None
 
         # COORS
-        self.norm_coor_weights = norm_coor_weights
-        self.norm_rel_coors = norm_rel_coors
-        if norm_rel_coors:
-            self.rel_coors_scale = nn.Parameter(torch.ones(1))
-
-        last_coor_linear = nn.Linear(m_dim * 4, 1)
         self.coors_mlp = nn.Sequential(
             nn.Linear(m_dim, m_dim * 4),
-            dropout,
+            self.dropout,
             SiLU(),
-            nn.Linear(m_dim * 4, 1)
+            nn.Linear(self.m_dim * 4, 1)
         ) if update_coors else None
-        # seems to be needed to keep the network from exploding to NaN with greater depths
-        self.init_eps = init_eps
+
         self.apply(self.init_)
 
     def init_(self, module):
         if type(module) in {nn.Linear}:
             # seems to be needed to keep the network from exploding to NaN with greater depths
-            nn.init.normal_(module.weight, std = self.init_eps)
+            nn.init.xavier_normal_(module.weight)
+            nn.init.zeros_(module.bias)
 
     def forward(self, x: Tensor, edge_index: Adj,
-                edge_attr: OptTensor = None, size: Size = None) -> Tensor:
+                edge_attr: OptTensor = None, batch: Adj = None, 
+                angle_data: List = None,  size: Size = None) -> Tensor:
         """ Inputs: 
             * x: (n_points, d) where d is pos_dims + feat_dims
+            * edge_index: (n_edges, 2)
             * edge_attr: tensor (n_edges, n_feats) excluding basic distance feats.
+            * batch: (n_points,) long tensor. specifies xloud belonging for each point
+            * angle_data: list of tensors (levels, n_edges_i, n_length_path) long tensor.
+            * size: None
         """
         coors, feats = x[:, :self.pos_dim], x[:, self.pos_dim:]
         
@@ -429,13 +594,13 @@ class EGNN_sparse(MessagePassing):
         # update coors if specified
         if self.update_coors:
             coor_wij = self.coors_mlp(m_ij)
+            # clamp if arg is set
+            if self.coor_weights_clamp_value:
+                coor_weights_clamp_value = self.coor_weights_clamp_value
+                coor_weights.clamp_(min = -clamp_value, max = clamp_value)
 
             # normalize if needed
-            if self.norm_rel_coors:
-                kwargs["rel_coors"] = F.normalize(kwargs["rel_coors"], dim = -1) * self.rel_coors_scale
-
-            if self.norm_coor_weights:
-                coor_wij = coor_wij.tanh()
+            kwargs["rel_coors"] = self.coors_norm(kwargs["rel_coors"])
 
             mhat_i = self.aggregate(coor_wij * kwargs["rel_coors"], **aggr_kwargs)
             coors_out = kwargs["coors"] + mhat_i
@@ -444,12 +609,12 @@ class EGNN_sparse(MessagePassing):
 
         # update feats if specified
         if self.update_feats:
-            if self.norm_feats:
-                hidden_feats = self.node_norm(kwargs["x"])
-            else:
-                hidden_feats = kwargs["x"]
-
+            # weight the edges if arg is passed
+            if self.soft_edge:
+                m_ij = m_ij * self.edge_weight(m_ij)
             m_i = self.aggregate(m_ij, **aggr_kwargs)
+
+            hidden_feats = self.node_norm(kwargs["x"])
             hidden_out = self.node_mlp( torch.cat([hidden_feats, m_i], dim = -1) )
             hidden_out = kwargs["x"] + hidden_out
         else: 
@@ -460,7 +625,7 @@ class EGNN_sparse(MessagePassing):
 
     def __repr__(self):
         dict_print = {}
-        return "E(n)-GNN Layer for Graphs " + str(dict_print) 
+        return "E(n)-GNN Layer for Graphs " + str(self.__dict__) 
 
 
 class EGNN_Sparse_Network(nn.Module):
@@ -483,15 +648,17 @@ class EGNN_Sparse_Network(nn.Module):
                                the resulting embedding. 1 entry per embedding needed. 
         * recalc: int. Recalculate edge feats every `recalc` MPNN layers. 0 for no recalc
         * verbose: bool. verbosity level.
+        -----
+        Diff with normal layer: one has to do preprocessing before (radius, global token, ...)
     """
     def __init__(self, n_layers, feats_dim, pos_dim = 3,
                        edge_attr_dim = 0, m_dim = 16,
                        fourier_features = 0,
                        embedding_nums=[], embedding_dims=[],
                        edge_embedding_nums=[], edge_embedding_dims=[],
-                       update_coors=True, update_feats=True, norm_feats=True,
-                       norm_rel_coors=False, norm_coor_weights=False,
-                       recalc=0 ):
+                       update_coors=True, update_feats=True, 
+                       norm_feats=True, norm_coors=False, 
+                       coor_weights_clamp_value=None, recalc=0 ):
         super().__init__()
 
         self.n_layers         = n_layers 
@@ -525,8 +692,7 @@ class EGNN_Sparse_Network(nn.Module):
         self.norm_feats       = norm_feats
         self.update_feats     = update_feats
         self.update_coors     = update_coors
-        self.norm_rel_coors   = norm_rel_coors
-        self.norm_coor_weights= norm_coor_weights
+        self.coor_weights_clamp_value = coor_weights_clamp_value
         self.recalc           = recalc
         
         # instantiate layers
@@ -537,10 +703,10 @@ class EGNN_Sparse_Network(nn.Module):
                                 m_dim = m_dim,
                                 fourier_features = fourier_features, 
                                 norm_feats = norm_feats,
+                                norm_coors = norm_coors,
                                 update_feats = update_feats,
-                                update_coors = update_coors,
-                                norm_rel_coors = norm_rel_coors,
-                                norm_coor_weights = norm_coor_weights)
+                                update_coors = update_coors, 
+                                coor_weights_clamp_value = coor_weights_clamp_value)
             self.mpnn_layers.append(layer)
 
     def forward(self, x, edge_index, batch, edge_attr,
@@ -559,11 +725,11 @@ class EGNN_Sparse_Network(nn.Module):
             
             # EDGES - Embedd each dim to its target dimensions:
             if edges_need_embedding:
-                edge_attr = embedd_token(x, self.edge_embedding_dims, self.edge_emb_layers)
+                edge_attr = embedd_token(edge_attr, self.edge_embedding_dims, self.edge_emb_layers)
                 edges_need_embedding = False
 
             # pass layers
-            x = layer(x, edge_index, edge_attr, size=bsize)
+            x = layer(x, edge_index, edge_attr, batch=batch, size=bsize)
 
             # recalculate edge info - not needed if last layer
             if self.recalc and ((i%self.recalc == 0) and not (i == len(self.mpnn_layers)-1)) :
